import re
from typing import List, Literal, Optional, Tuple

from . import selectors, utils


def parse_selection_expressions(expressions: List[str]) -> List[selectors.Selector]:
    selectors = []
    for index, expression in enumerate(expressions):
        try:
            selector = Parser.parse(expression)
        except SelectionExpressionError as error:
            raise ValueError(f"Invalid selection expression: {expression} at index {0} in {expressions}") from error
        selectors.append(selector)

    return selectors


class Parser:
    _ELEMENT_TEMPLATE = r'[0-9]{1,6}'
    # using fr, backslashes do not have to be escaped in f-strings
    _LIST_TEMPLATE = fr"^\[(?P<elements>{_ELEMENT_TEMPLATE}(,{_ELEMENT_TEMPLATE})*)\]$"
    _RANGE_TEMPLATE = f"^(?P<start>{_ELEMENT_TEMPLATE})?:(?P<stop>{_ELEMENT_TEMPLATE})?" + \
                      f"(:(?P<step>-?{_ELEMENT_TEMPLATE})?)?$"

    _LIST_PATTERN = re.compile(_LIST_TEMPLATE)
    _RANGE_PATTERN = re.compile(_RANGE_TEMPLATE)

    @staticmethod
    def _single_element_expression(
            expression: str,
            kind: Literal['sequence', 'frame']) -> Optional[selectors.SingleElementSelector]:
        max_length = 4 if kind == 'sequence' else 6

        if expression.isdigit() and len(expression) <= max_length:
            normalized_element = utils.normalize_element(expression, kind)

            return selectors.SingleElementSelector(expression, normalized_element)

        return None

    @staticmethod
    def _list_expression(expression: str, kind: Literal['sequence', 'frame']) -> Optional[selectors.ListSelector]:
        if match := Parser._LIST_PATTERN.match(expression):
            elements = match.group('elements').split(',')

            normalized_elements = [utils.normalize_element(element, kind) for element in elements]

            return selectors.ListSelector(expression, normalized_elements)

        return None

    @staticmethod
    def _star_expression(expression: str, _kind) -> Optional[selectors.StarSelector]:
        # star expression work the same for sequences and frames, therefore _kind is ignored
        # _kind is only specified to conform with the common signature
        return selectors.StarSelector() if expression == '*' else None

    @staticmethod
    def _data_expression(expression: str, kind: str) -> Optional[selectors.DataSelector]:
        if kind != 'sequence':
            raise ValueError(f"kind must be 'sequence': {kind}")

        return selectors.DataSelector() if expression == 'data' else None

    @staticmethod
    def _data_syn_expression(expression: str, kind: str) -> Optional[selectors.DataSynSelector]:
        if kind != 'sequence':
            raise ValueError(f"kind must be 'sequence': {kind}")

        return selectors.DataSynSelector() if expression == 'data_syn' else None

    @staticmethod
    def _range_expression(expression: str, kind: Literal['sequence', 'frame']) -> Optional[selectors.RangeSelector]:
        if match := Parser._RANGE_PATTERN.match(expression):
            start = utils.normalize_optional_element(match.group('start'), kind)
            stop = utils.normalize_optional_element(match.group('stop'), kind)
            step = int(match.group('step') or 1)

            return selectors.RangeSelector(expression, start, stop, step)

    @staticmethod
    def _get_sequence_and_frame_expression(expression: str) -> Tuple[str, str]:
        if expression.count('/') == 0:
            raise SelectionExpressionError(f"Expression contains no '/': {expression}", expression)
        if expression.count('/') > 1:
            raise SelectionExpressionError(f"Expression must contain only one '/': {expression}", expression)

        sequence_expression, frame_expression = expression.split('/')

        if not sequence_expression:
            raise SelectionExpressionError(f"No Sequence Expression given: {expression}", expression)
        if not frame_expression:
            raise SelectionExpressionError(f"No Frame Expression given: {expression}", expression)

        return sequence_expression, frame_expression

    @staticmethod
    def _parse_sequence_expression(expression: str):
        for expression_handler in [Parser._single_element_expression,
                                   Parser._list_expression,
                                   Parser._star_expression,
                                   Parser._data_expression,
                                   Parser._data_syn_expression,
                                   Parser._range_expression]:
            if selector := expression_handler(expression, 'sequence'):
                return selector

        # no expression handler matched
        raise SelectionExpressionError(f"Invalid sequence selection expression: {expression}", expression)

    @staticmethod
    def _parse_frame_expression(expression: str):
        for expression_handler in [Parser._single_element_expression,
                                   Parser._list_expression,
                                   Parser._star_expression,
                                   Parser._range_expression]:
            if selector := expression_handler(expression, 'frame'):
                return selector

        # no expression handler matched
        raise SelectionExpressionError(f"Invalid frame selection expression: {expression}", expression)

    @staticmethod
    def parse(expression: str):
        sequence_expression, frame_expression = Parser._get_sequence_and_frame_expression(expression)

        try:
            sequence_selector = Parser._parse_sequence_expression(sequence_expression)
            frame_selector = Parser._parse_frame_expression(frame_expression)
        except Exception as error:
            raise SelectionExpressionError(
                f"Selection expression could not be parsed: {expression}",
                expression) from error

        return selectors.Selector(expression, sequence_selector, frame_selector)


class SelectionExpressionError(Exception):
    def __init__(self, message: str, expression: str):
        super().__init__(message)
        self.expression = expression



import re
import sys

from . import info
from .loader import Loader


def print_usage_info():
    print('usage: ycbvideo')
    print('                DATASET_ROOT [-v | -vv]')


# sys.argv[0] is the name of the file executed
if (argument_count := len(sys.argv)) >= 2:
    dataset = sys.argv[1]
    loader = Loader(dataset)
    verbosity = 0
    frame_sequences = (f"{index:04d}" for index in range(92))

    if len(sys.argv) > 2:
        # verbosity level -v == 0 is the default
        if match := re.match(r'^-(?P<vstring>[v]+)$', sys.argv[2]):
            verbosity = len(match.group('vstring'))

    info.print_frame_info(loader, frame_sequences, verbosity)
else:
    print_usage_info()



import functools
import os
from pathlib import Path
import random
from typing import List, Iterable, Union, Dict, Optional, Sequence

from . import datatypes, parsing, selectors, utils


def _load_selection_expressions_from_file(file: Union[Path, str]) -> List[str]:
    file_path = Path(file)

    if not file_path.exists():
        raise IOError(f"File does not exist: {file_path}")

    expressions = []
    with open(file_path, 'r') as f:
        for line_number, line in enumerate(f):
            # skip empty lines and remove leading and trailing whitespace including 'newline'
            if stripped_line := line.lstrip().rstrip():
                expressions.append(stripped_line)

    return expressions


class Loader:
    def __init__(self, path: Union[Path, str]):
        self._path = utils.validate_directory_path(path)
        self._data_directory = self._path / 'data'
        self._data_syn_directory = self._path / 'data_syn'
        self._available_frame_sequences = self.get_available_frame_sequences()

    def frames_info(self) -> Dict[str, Dict[str, Optional[List[str]]]]:
        available_sequences = sorted(self.get_available_frame_sequences())

        info = {}
        for sequence_identifier in available_sequences:
            sequence = self.get_frame_sequence(sequence_identifier)
            complete_frame_sets, incomplete_frame_sets = sequence.get_available_frame_sets()

            frame_sets = {}
            for complete_set in complete_frame_sets:
                frame_sets[complete_set] = None
            for identifier, incomplete_set in incomplete_frame_sets.items():
                frame_sets[identifier] = incomplete_set

            info[sequence_identifier] = frame_sets

        return info

    def get_available_data_frame_sequences(self) -> List[str]:
        return os.listdir(self._data_directory)

    def get_available_frame_sequences(self) -> List[str]:
        available_sequences = self.get_available_data_frame_sequences()

        if self._data_syn_directory.exists():
            available_sequences.append(self._data_syn_directory.name)

        return available_sequences

    @functools.lru_cache()
    def get_frame_sequence(self, index: Union[str, int]) -> datatypes.FrameSequence:
        if isinstance(index, int):
            index = "{:04d}".format(index)

        path = self._data_syn_directory if index == self._data_syn_directory.name else self._data_directory / index

        if index not in self.get_available_frame_sequences():
            raise IOError(f"Frame sequence does not exist: {index} ({path})")

        return datatypes.FrameSequence(path)

    def _get_descriptors(self, expressions: List[str]) -> List[datatypes.Descriptor]:
        selectors = parsing.parse_selection_expressions(expressions)

        descriptors = []
        for selector in selectors:
            descriptors.extend(self._create_descriptors_from_selector(selector))

        return descriptors

    def _create_descriptors_from_selector(self, selector: selectors.Selector) -> List[datatypes.Descriptor]:
        available_sequences = sorted(self.get_available_frame_sequences())

        descriptors = []
        try:
            selected_sequences = selector.select_sequences(available_sequences)
        except selectors.MissingElementError as error:
            raise IOError(f"Frame sequence is not available: {error.element}")

        for sequence in selected_sequences:
            sequence_object = self.get_frame_sequence(sequence)
            available_frames = sorted(sequence_object.get_complete_frame_sets())

            try:
                selected_frames = selector.select_frames(available_frames)
            except selectors.MissingElementError as error:
                frame = error.element
                if frame in (incomplete_frame_sets := sequence_object.get_incomplete_frame_sets()):
                    missing_files = incomplete_frame_sets[frame]

                    raise IOError(f"Files for frame missing: {sequence}/{frame} misses {missing_files}")
                else:
                    raise IOError(f"Frame is not available: {sequence}/{frame}")

            for frame in selected_frames:
                descriptors.append(datatypes.Descriptor(sequence, frame))

        return descriptors

    def frames(self, frames: Union[List[str], Union[Path, str]], shuffle: bool = False) -> Iterable[datatypes.Frame]:
        """
        Return an iterable object for the specified frames.

        To be memory-friendly, the specified frames will be created
        just-in-time. The frames can be specified as either a list of
        selection expressions or as a path to a file with one selection
        expression per line.
        Frame sequences and frames can each be selected by an
        expression containing
            - a single element
            - a list of comma-delimited elements
            - a range of elements similar to slicing works in Python
            - a '*' expressing all available elements.
        Also, 'data_syn' can be used as a frame sequence selection
        element to specify the data_syn frame sequence.
        To get all available frame sequences except for data_syn,
        'data' can be specified as a frame sequence selection
        element.
        A frame sequence selection expression and a frame selection
        expression together form a selection expression. Between both
        frame sequence selection expression and frame selection
        expression, a '/' is placed as the delimiter.

        Examples for valid selection expressions:
            - '1/42'
                Get the 42th frame from the 1st frame sequence
            - '1/[1,2,3]'
                Get the 1st, 2nd and 3th frame from the 1st frame
                sequence
            - '[1,2]/42'
                Get the 42th frame from the 1st and 2nd frame
                sequence each
            - '[1,2]/[4,3]'
                Get the 4th and 3rd frame from the 1st and 2nd frame
                sequence each
            - '[1,2]/*'
                Get all available frames from the 1st and 2nd frame
                sequence each
            - '*/[3,4]'
                Get the 3rd and 4th frame from each the available
                frame sequences
            - '*/*'
                Get all available frames from each the available
                frame sequences
            - 'data_syn/[1,2]'
                Get the 1st and 2nd frame from the data_syn frame
                sequence
            - 42:56:1/1
                Get the 1st frame from each frame sequence between frame
                sequence 42 (inclusive) and 56 (exclusive) by taking steps
                of length 1, i.e. when frame sequence 42 to frame sequence
                55 all are available, 14 frames would be selected
            - 42/::-2
                Get every other frame from frame sequence 42 in reverse
                order from all available frames of frame sequence 42

        If `shuffle` is set to 'False', the frames will be returned
        in exactly the order as specified by the order of selection
        expressions and the order of the elements in those. Otherwise,
        the frames get shuffled. Since random.shuffle() is used to
        shuffle the frames, by setting a seed with random.seed() or
        setting a state with random.setstate() a certain shuffle order
        can be defined.

        For each specified frame it is made sure that the corresponding
        files exists, even if the frame is only eventually created.
        This can be especially useful when specifying a large number of
        frames and consuming them needs minutes or even hours e.g. in
        a machine-learning context.

        Parameters
        ----------
        frames : Union[List[str], Union[pathlib.Path, str]]
            Either a list of selection expressions or a path to a file
            containing selection expressions.
            The path can be either a string or a Path object.
            If path is not absolute, it is assumed to be relative to
            the root of the dataset.
        shuffle : bool, optional
            If True, the selected frames get shuffled

        Returns
        ------
        Iterable[Frame]
            An iterable allowing access to the frame objects with
            additional support for the builtin len(iterable) and
            for accessing single frame objects from it by specifying
            an int index e.g. iterable[42]
        """

        if isinstance(frames, list):
            if not frames:
                raise ValueError('List is empty')
            frame_descriptors = self._get_descriptors(frames)
        elif isinstance(frames, (Path, str)):
            path = Path(frames)

            if not path.is_absolute():
                path = self._path / path

            if not path.exists():
                raise IOError(f"Path does not exist: {path}")
            elif not path.is_file():
                raise IOError(f"Path is not a file: {path}")

            expressions = _load_selection_expressions_from_file(path)
            frame_descriptors = self._get_descriptors(expressions)
        else:
            raise TypeError('frames has to be of type list, pathlib.Path or str')

        if shuffle:
            random.shuffle(frame_descriptors)

        return _FrameAccessor(self, frame_descriptors)


class _FrameAccessor:
    def __init__(self, loader: Loader, descriptors: Sequence[datatypes.Descriptor]):
        self._loader = loader
        self._descriptors = descriptors

    def _get_frame(self, descriptor: datatypes.Descriptor) -> datatypes.Frame:
        sequence_object = self._loader.get_frame_sequence(descriptor.frame_sequence)

        return sequence_object.get_frame(descriptor.frame)

    def __iter__(self):
        for descriptor in self._descriptors:
            yield self._get_frame(descriptor)

    def __len__(self):
        return len(self._descriptors)

    @functools.lru_cache()
    def __getitem__(self, index: int):
        if isinstance(index, int):
            if 0 <= index < len(self):
                return self._get_frame(self._descriptors[index])

            raise ValueError('index out of range')

        raise TypeError('item must be an int')



import pkgutil

from .loader import Loader
from .datatypes import Frame, Descriptor, Box

__version__ = pkgutil.get_data(__package__, 'VERSION').decode('utf-8').rstrip()

__all__ = [
    'Loader',
    'Frame',
    'Descriptor',
    'Box'
]



import functools
import os
from pathlib import Path
import re
from typing import Union, NamedTuple, Tuple, List, Dict, Optional

import imageio
from numpy import ndarray

from . import utils

FILE_PATTERN = re.compile(r'^(?P<index>[0-9]{6})-(?P<kindoffile>[a-z]+)\.(png|txt)$')
BOUNDING_BOX_PATTERN = re.compile(r'^(?P<label>[^ ]+) ([0-9.]+) ([0-9.]+) ([0-9.]+) ([0-9.]+)$')


class Box(NamedTuple):
    label: str
    coordinates: Tuple[float, float, float, float]


class Descriptor(NamedTuple):
    frame_sequence: str
    frame: str


class Frame(NamedTuple):
    color: ndarray
    depth: ndarray
    boxes: Optional[List[Box]]
    label: ndarray
    description: Descriptor


class FrameSequence:
    def __init__(self, path: Union[Path, str]):
        self._path = utils.validate_directory_path(path)
        self._sequence_name = self._path.name

    @functools.lru_cache()
    def get_available_frame_sets(self) -> Tuple[List[str], Dict[str, List[str]]]:
        files = os.listdir(self._path)

        frame_sets = {}
        for file in files:
            if match := FILE_PATTERN.match(file):
                index = match.group('index')
                kind_of_file = match.group('kindoffile')

                if index not in frame_sets:
                    frame_sets[index] = set()

                frame_sets[index].add(kind_of_file)

        expected_kind_of_files = ['color', 'depth', 'label']
        if self._sequence_name != 'data_syn':
            expected_kind_of_files.append('box')

        complete_frame_sets = []
        incomplete_frame_sets = {}
        for frame_set, kind_of_files in frame_sets.items():
            missing_kinds_of_file = []

            for kind_of_file in expected_kind_of_files:
                if kind_of_file not in kind_of_files:
                    missing_kinds_of_file.append(kind_of_file)

            if missing_kinds_of_file:
                incomplete_frame_sets[frame_set] = missing_kinds_of_file
            else:
                complete_frame_sets.append(frame_set)

        return complete_frame_sets, incomplete_frame_sets

    def get_complete_frame_sets(self) -> List[str]:
        complete_frame_sets, _ = self.get_available_frame_sets()

        return complete_frame_sets

    def get_incomplete_frame_sets(self) -> Dict[str, List[str]]:
        _, incomplete_frame_sets = self.get_available_frame_sets()

        return incomplete_frame_sets

    def __len__(self):
        return len(self.get_complete_frame_sets())

    def get_frame(self, index: Union[int, str]) -> Frame:
        if isinstance(index, int):
            index = "{:06d}".format(index)

        if index in (incomplete_frame_sets := self.get_incomplete_frame_sets()):
            raise IOError(f"Frame set is not complete: {index} Missing: {incomplete_frame_sets[index]}")
        if index not in self.get_complete_frame_sets():
            raise IOError(f"Frame set does not exist: {index}")

        return Frame(
            color=imageio.imread(self._path / f"{index}-color.png"),
            depth=imageio.imread(self._path / f"{index}-depth.png"),
            boxes=self._get_boxes(index) if self._sequence_name != 'data_syn' else None,
            label=imageio.imread(self._path / f"{index}-label.png"),
            description=Descriptor(self._path.name, index))

    def _get_boxes(self, index: str) -> List[Box]:
        file = self._path / f"{index}-box.txt"
        with open(file, 'r') as f:
            boxes = []

            for line in f:
                if match := BOUNDING_BOX_PATTERN.match(line):
                    label = match.group(1)
                    coordinates = (float(match.group(2)),
                                   float(match.group(3)),
                                   float(match.group(4)),
                                   float(match.group(5)))
                    boxes.append(Box(label, coordinates))
                else:
                    raise ValueError(f"Box file has invalid format: {file}")

        return boxes



from pathlib import Path
from typing import Union, Optional, Literal


def validate_directory_path(path: Union[Path, str]) -> Path:
    path = Path(path)
    if not path.exists():
        raise IOError(f"Path does not exist: {path}")
    if not path.is_dir():
        raise IOError(f"Path is not a directory: {path}")

    return path


def check_kind(kind: str):
    if kind not in ['sequence', 'frame']:
        raise ValueError(f"kind has to be either 'sequence' or 'frame': {kind}")


def normalize_element(element: str, kind: Literal['sequence', 'frame']) -> str:
    check_kind(kind)

    if not element.isdigit():
        raise ValueError(f"element must be a digit: {element}")

    if kind == 'sequence' and len(element) > 4:
        raise ValueError(f"sequence element must be of length <= 4: {element}")
    elif kind == 'frame' and len(element) > 6:
        raise ValueError(f"frame element must be of length <= 6: {element}")

    if kind == 'sequence':
        return "{:04d}".format(int(element))
    else:
        return "{:06d}".format(int(element))


def normalize_optional_element(element: Optional[str], kind: Literal['sequence', 'frame']) -> Optional[str]:
    if not element:
        return element

    return normalize_element(element, kind)



from abc import ABC, abstractmethod
from typing import List, Optional


class ElementSelector(ABC):
    @abstractmethod
    def select(self, elements: List[str]) -> List[str]:
        raise NotImplementedError

    @staticmethod
    def check_elements_to_select_from(elements: List[str]):
        if not elements:
            raise ValueError("No elements to select from")


class SingleElementSelector(ElementSelector):
    def __init__(self, expression: str, element: str):
        self._expression = expression
        self._element_to_select = element

    def select(self, elements: List[str]) -> List[str]:
        ElementSelector.check_elements_to_select_from(elements)

        if self._element_to_select not in elements:
            raise MissingElementError(
                message=f"Single element missing: {self._expression}",
                element=self._element_to_select,
                elements=elements)

        return [self._element_to_select]


class ListSelector(ElementSelector):
    def __init__(self, expression: str, elements: List[str]):
        self._expression = expression
        self._elements_to_select = elements

    def select(self, elements: List[str]) -> List[str]:
        ElementSelector.check_elements_to_select_from(elements)

        for index, element in enumerate(self._elements_to_select):
            if element not in elements:
                raise MissingElementError(
                    message=f"List element missing: {self.get_list_expression_element(index)} at index {index}",
                    element=element,
                    elements=elements)

        return self._elements_to_select

    def get_list_expression_element(self, index: int) -> str:
        return self._expression.lstrip('[').rstrip(']').split(',')[index]


class StarSelector(ElementSelector):
    def select(self, elements: List[str]) -> List[str]:
        ElementSelector.check_elements_to_select_from(elements)

        return elements


class DataSelector(ElementSelector):
    def select(self, elements: List[str]) -> List[str]:
        ElementSelector.check_elements_to_select_from(elements)

        return [element for element in elements if element != 'data_syn']


class DataSynSelector(ElementSelector):
    def select(self, elements: List[str]) -> List[str]:
        ElementSelector.check_elements_to_select_from(elements)

        if 'data_syn' not in elements:
            raise MissingElementError(
                message="'data_syn' sequence missing",
                element='data_syn',
                elements=elements
            )

        return ['data_syn']


class RangeSelector(ElementSelector):
    def __init__(self, expression: str, start: Optional[str], stop: Optional[str], step: int = 1):
        self._expression = expression

        self.check_range_elements(start, stop, step)
        self._start = start
        self._stop = stop
        self._step = step

    def select(self, elements: List[str]) -> List[str]:
        ElementSelector.check_elements_to_select_from(elements)

        if self._start:
            try:
                start_index = elements.index(self._start)
            except ValueError as error:
                raise MissingElementError(
                    f"Range start element missing: {self._expression}", self._start, elements) from error
        else:
            start_index = None

        if self._stop:
            try:
                stop_index = elements.index(self._stop)
            except ValueError as error:
                raise MissingElementError(f"Range stop element missing: {self._stop}", self._stop, elements) from error
        else:
            stop_index = None

        expansion = elements[slice(start_index, stop_index, self._step)]

        # an empty expansion is of no use and would be the result of
        # start == stop
        # start = None, stop == first element and step size > 1
        # start = last element, stop == last element and step size < -1
        if not expansion:
            raise EmptySelectionError(f"No elements selected: {self._expression}", self._expression, elements)

        return expansion

    def check_range_elements(self, start: Optional[str], stop: Optional[str], step: int):
        """
        Check for correct values for start, stop and step.

        Cases, where "expanding" the range, i.e. selecting all elements which
        lay in the range borders, results in an empty selection are not
        handled here since not all of them could be detected at this point.
        'start == stop' (resulting in an empty expansion) is an example,
        which could be detected at this point, but cases like
        'start == None and stop equals the first available element'
        can only be detected while expanding.
        """
        if step == 0:
            raise ValueError(f"step must not be 0")

        if start and stop:
            if step > 0 and int(start) > int(stop):
                raise ValueError(f"step > 0 requires start < stop: {self._expression}")

            if step < 0 and int(start) < int(stop):
                raise ValueError(f"step < 0 requires start > stop: {self._expression}")


class Selector:
    def __init__(self, expression: str, sequence_selector: ElementSelector, frame_selector: ElementSelector):
        self._expression = expression
        self._sequence_selector = sequence_selector
        self._frame_selector = frame_selector

    def select_sequences(self, sequences: List[str]) -> List[str]:
        return self._sequence_selector.select(sequences)

    def select_frames(self, frames: List[str]) -> List[str]:
        return self._frame_selector.select(frames)


class SelectionError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class MissingElementError(SelectionError):
    def __init__(self, message: str, element: str, elements: List[str]):
        super().__init__(message)
        self.element = element
        self.elements = elements


class EmptySelectionError(SelectionError):
    def __init__(self, message: str, expression: str, elements: List[str]):
        super().__init__(message)
        self.expression = expression
        self.elements = elements
